'use strict';

var Describer = (function () {
    function Describer() {
    }
    Describer.getParameters = function (instance) {
        return Object.getOwnPropertyNames(instance).filter(function (propertyName) { return !(instance[propertyName] instanceof Function); });
    };
    return Describer;
}());

var BuildConfiguration = (function () {
    function BuildConfiguration() {
        this.required = new Set;
        this.ignores = new Set;
        this.transformers = {};
        this.aliases = {};
        this.listElementConstructors = {};
    }
    BuildConfiguration.prototype.require = function () {
        var _this = this;
        var paths = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            paths[_i] = arguments[_i];
        }
        paths.forEach(function (path) { return _this.required.add(path); });
        return this;
    };
    BuildConfiguration.prototype.ignore = function () {
        var _this = this;
        var paths = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            paths[_i] = arguments[_i];
        }
        paths.forEach(function (path) { return _this.ignores.add(path); });
        return this;
    };
    BuildConfiguration.prototype.transform = function (localPath, transformerIn, transformerOut) {
        var defaultTransformer = function (value) { return value; };
        this.transformers[localPath] = {
            "in": transformerIn,
            out: transformerOut !== null && transformerOut !== void 0 ? transformerOut : defaultTransformer
        };
        return this;
    };
    BuildConfiguration.prototype.alias = function (foreignPath, localPath) {
        this.aliases[localPath] = foreignPath;
        return this;
    };
    BuildConfiguration.prototype.listType = function (localPath, builtObject) {
        this.listElementConstructors[localPath] = builtObject;
        return this;
    };
    return BuildConfiguration;
}());

/**
 * performs deep clone. Not able to deep clone closures, enclosed values
 * still point to original reference, so this is not full deep clone in all situations.
 * Anyway it is able to detect circular references
 */

function clone(x) {
  if (isArray(x)) {
    return cloneArray(x);
  } else if (isFunction(x)) {
    return x;
  }else if(isObject(x)) {
    return cloneObject(x);
  }else {
    return x;
  }

}

/**
 * objects constructed by other functions than Object, because those cannot be
 simply copied as simple object
 */
function isObject(o) {
  //typeof o === "object" inclues arrays and everything else, we want to target only objec-objects here
  return Object.prototype.toString.call(o) === "[object Object]";
}
function isArray(a) {
  return Object.prototype.toString.call(a) === '[object Array]';
}

/**
 * detects just pure functions
 */
function isFunction(f){
  if(f instanceof Function) return true;
  return false;
}

function cloneArray(arr) {
  var cloned = arr.map(function(member) {
    return clone(member);
  });
  return cloned;
}

/**
 * @param obj object to clone
 * @param sources if recursive, allready cloned objects in order to avoid circular references
 * @param clones allready cloned source objects, for avoiding circular references
 */
function cloneObject(obj, sources, clones) {
  if (!obj) return obj;
  if (obj.clone && isFunction(obj.clone)) return obj.clone();
  if(obj instanceof Date) {
    //date is built in type and can be created only by invoking new Date()
    //https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date
    //there are more such objects as Error, but they are ignored here
    return new Date(obj.getTime());
  }
  sources = sources || [];
  clones = clones || [];
  var target = Object.create(obj);
  var properties = Object.getOwnPropertyNames(obj);
    properties.forEach( function(property) {
      var descriptor = Object.getOwnPropertyDescriptor(obj, property);
      descriptor.value = _getValue2Assign(descriptor.value, obj, target, sources, clones);
      Object.defineProperty(target, property, descriptor);
    });
  return target;
}


function _getValue2Assign(value, objectBeingCloned, target, sources, clones) {
  if (!isObject(value)){
    return clone(value);
  }else if(value.clone) {//cloneable
    var target = value.clone();
    if (target && value.constructor.prototype.isPrototypeOf(clone.constructor.prototype)) {
      return target;
    }
    return _getObjectValue2Assign(value, objectBeingCloned, target, sources, clones);
  }
  //else we deal with new object
  return _getObjectValue2Assign(value, objectBeingCloned, target, sources, clones);
}

function _getObjectValue2Assign(value, objectBeingCloned, target, sources, clones) {
  var allreadyClonedIndex = sources.indexOf(value);
  if(allreadyClonedIndex !== -1) {
    return clones[allreadyClonedIndex];
  }else {
    sources.push(objectBeingCloned);
    clones.push(target);
    return cloneObject(value, sources, clones);
  }
}

var deepclone=clone;

function deepCopy(source) {
    return deepclone(source);
}

var BuildableResource = (function () {
    function BuildableResource() {
    }
    Object.defineProperty(BuildableResource.prototype, "buildConfig", {
        get: function () {
            if (!Object.getOwnPropertyDescriptor(this, 'resourceBuildConfiguration') && this.resourceBuildConfiguration) {
                this.resourceBuildConfiguration = deepCopy(this.resourceBuildConfiguration);
            }
            else if (!this.resourceBuildConfiguration) {
                this.resourceBuildConfiguration = new BuildConfiguration;
            }
            return this.resourceBuildConfiguration;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(BuildableResource.prototype, "build", {
        get: function () {
            var c = this.constructor.prototype.constructor;
            return new Builder(c, this.buildConfig);
        },
        enumerable: false,
        configurable: true
    });
    BuildableResource.prototype.fromJSON = function (source) {
        return this.build.fromJSON(source);
    };
    BuildableResource.prototype.toJSON = function () {
        return this.build.toJSON(this);
    };
    return BuildableResource;
}());

var Dot = (function () {
    function Dot() {
    }
    Dot.get = function (path, object, separator) {
        if (separator === void 0) { separator = '.'; }
        return path
            .split(separator)
            .reduce(function (level, key) {
            if (level)
                return level[key];
        }, object);
    };
    Dot.assign = function (path, object, value, separator) {
        if (separator === void 0) { separator = '.'; }
        var levels = path.split(separator);
        if (levels.length === 0)
            throw new Error("Please provide a non-empty path.");
        else if (levels.length === 1)
            object[levels[0]] = value;
        else {
            if (object[levels[0]])
                return Dot.assign(levels.slice(1).join(separator), object[levels[0]], value);
            else {
                object[levels[0]] = {};
                return Dot.assign(levels.slice(1).join(separator), object[levels[0]], value);
            }
        }
    };
    return Dot;
}());

var Builder = (function () {
    function Builder(ctor, buildConfig) {
        if (buildConfig === void 0) { buildConfig = null; }
        var _a;
        this.baseObject = new ctor();
        this.buildConfig = (_a = buildConfig !== null && buildConfig !== void 0 ? buildConfig : this.baseObject.buildConfig) !== null && _a !== void 0 ? _a : new BuildConfiguration;
    }
    Builder.prototype.getForeignObjectReference = function (localPath, foreignObject) {
        var _a;
        if (foreignObject === void 0) { foreignObject = undefined; }
        var foreignPath = (_a = this.buildConfig.aliases[localPath]) !== null && _a !== void 0 ? _a : localPath;
        var foreignValue = Dot.get(foreignPath, foreignObject);
        var result;
        if (foreignObject == undefined) {
            result = { path: foreignPath, value: null };
        }
        else {
            result = { path: foreignValue != null ? foreignPath : localPath, value: foreignValue };
        }
        result.value = Dot.get(result.path, foreignObject);
        return result;
    };
    Builder.prototype.fromJSON = function (source) {
        var _this = this;
        var target = this.baseObject;
        var params = Describer.getParameters(target);
        params.forEach(function (param) {
            if (_this.buildConfig.ignores.has(param) || !params.includes(param)) {
                return;
            }
            var foreignObject = _this.getForeignObjectReference(param, source);
            if (!source || !foreignObject.value) {
                if (_this.buildConfig.required.has(param)) {
                    throw new Error("Invalid input object, missing required parameter: " + param);
                }
                else
                    return;
            }
            if (target[param] instanceof BuildableResource) {
                target[param] = target[param].build.fromJSON(foreignObject.value);
            }
            else if (Array.isArray(target[param])) {
                var list = foreignObject.value;
                target[param] = list.map(function (item, index) {
                    var listClassElement = _this.buildConfig.listElementConstructors[param];
                    if (listClassElement) {
                        var listElementClassBuilder = listClassElement.build;
                        return listElementClassBuilder.fromJSON(item);
                    }
                    else if (!_this.buildConfig.required.has(param) && foreignObject.value) {
                        return foreignObject.value[index];
                    }
                });
            }
            else {
                if (_this.buildConfig.transformers[param]) {
                    target[param] = _this.buildConfig.transformers[param]["in"](foreignObject.value);
                }
                else {
                    target[param] = foreignObject.value;
                }
            }
        });
        return deepCopy(target);
    };
    Builder.prototype.toJSON = function (source) {
        var _this = this;
        var params = Describer.getParameters(source);
        var result = {};
        params.forEach(function (param) {
            var foreignObject = _this.getForeignObjectReference(param);
            var valueToAssign = undefined;
            if (_this.buildConfig.ignores.has(param)) {
                return;
            }
            if (source[param] instanceof BuildableResource) {
                valueToAssign = source[param].build.toJSON(source[param]);
            }
            else if (Array.isArray(source[param])) {
                var list = source[param];
                valueToAssign = list.map(function (item, index) {
                    var listClassElement = _this.buildConfig.listElementConstructors[param];
                    if (listClassElement) {
                        var listElementClassBuilder = listClassElement.build;
                        return listElementClassBuilder.toJSON(item);
                    }
                    else if (source.hasOwnProperty(param)) {
                        var x = Dot.get(foreignObject.path, result);
                        return _this.buildConfig.transformers[param] ? _this.buildConfig.transformers[param].out(x[index]) : x[index];
                    }
                });
            }
            else {
                valueToAssign = _this.buildConfig.transformers[param] ? _this.buildConfig.transformers[param].out(source[param]) : source[param];
            }
            Dot.assign(foreignObject.path, result, valueToAssign);
        });
        delete result["resourceBuildConfiguration"];
        return result;
    };
    return Builder;
}());

exports.BuildableResource = BuildableResource;
exports.Builder = Builder;
